#include "platform.h"
#include <stdio.h>
#include <conio.h>
#include "comm.h"
#include "optionparse.h"
#include "error.h"
#include "lpt.h"

CommBase* g_pComm = nullptr; // LPT only atm. USB is a separate tool.

static const char versionString[] = "0.9D"; // Note: puting this in sync w/ bootloader ROMs. Prior 1.x versions were never public.

enum
{
	COMMAND_NOP        = 0,
	COMMAND_REBOOTRAM  = 1,
	COMMAND_REBOOTROM  = 2,
	COMMAND_UPLOADMAIN = 3,
	COMMAND_UPLOADSUB  = 4,
};

bool Nop ()       { return g_pComm->SendByte (COMMAND_NOP); }
bool RebootRAM () { return g_pComm->SendByte (COMMAND_REBOOTRAM); }
bool RebootROM () { return g_pComm->SendByte (COMMAND_REBOOTROM); }

// Write a 16 bit value in high-low (bigendian) order.
static bool COMM_SendWord (uint16 word)
{
	return g_pComm->SendByte (word >> 8) && g_pComm->SendByte (word & 0xff);
}

static bool UploadInternal (const uint8* pData, uint16 nBytes)
{
	// Command has been sent.
	if (!COMM_SendWord (nBytes))
		return false;

	// Now send out all the data.
	for (uint32 i=0; i<nBytes; i++)
		if (!g_pComm->SendByte(pData[i]))
			return false;

	return true;
}

bool UploadMainRam (const void* pData, uint16 nBytes)
{
	DEBUG_ASSERT(pData);
	DEBUG_ASSERT(nBytes <= 32768);
	if (!g_pComm->SendByte (COMMAND_UPLOADMAIN))
		return false;

	return UploadInternal ((const uint8*)pData, nBytes);
}

bool UploadSubRam (const void* pData, uint16 nBytes)
{
	DEBUG_ASSERT(pData);
	DEBUG_ASSERT(nBytes <= 32768);
	if (!g_pComm->SendByte (COMMAND_UPLOADSUB))
		return false;

	return UploadInternal ((const uint8*)pData, nBytes);
}

// Options:
// -port <lpt port>; default 0x378.
// -txdelay <ns>; default 50000.
// -txtimeout <ms>; default 0 (disabled).
// -debugdelay (ms); default 0 (disabled).
// -console; reads debug output instead of exiting.

static bool s_bConsole = false;
static bool SetConsole (const char*, void*) { s_bConsole = true; return true; }

// Note:
// Serial support (unreleased) has been stripped out. This was of little use, and orboot-usb doesn't use serial (anymore).
// In fact, there's very little overlap (parallel logic is done on the OutRun-USB board itself), so the tool is entirely separate.
//static uint32 s_serialPort = 0;

static CommLPT::Options s_lptOptions;
static bool s_lptPortDefined = false;
static bool SetPort (const char* portStr, void*) 
{
	uint32 port;
	if (sscanf (portStr, "0x%x", &port) == 1 || 
		sscanf (portStr, "0x%X", &port) == 1 || 
		sscanf (portStr, "%u",   &port) == 1)
	{
		if (port == 378) 
			printwarn ("Warning: did you mean to use hexadecimal (0x378) notation?\n");
		s_lptOptions.basePort = port;
		s_lptPortDefined = true;
		return true;
	}
	/*
	else if ((sscanf (portStr, "COM%u", &port) == 1 || sscanf (portStr, "com%u", &port) == 1) && port > 0)
	{
		// Serial.
		s_serialPort = port;
		return true;
	}
	*/
	else
	{
		printerr ("Port option needs a valid positive value for LPT!\n");
		return false;
	}
}

static bool SetTxDelayLPT (const char* delayStr, void*)
{
	uint64 txDelay;
	if (sscanf (delayStr, "%I64u", &txDelay) == 1)
	{
		s_lptOptions.txDelayNS = txDelay;
		if (txDelay < 25000)
			printwarn ("Warning: TX delay should be at least 35000 nanoseconds.\n");
		return true;
	}
	else 
	{
		printerr ("TX delay option needs a valid positive value!\n");
		return false;
	}
}

static bool ParseInt32 (const char* str, uint32& dest)
{
	uint32 tmp;
	if (sscanf (str, "%u", &tmp) == 1)
	{
		dest = tmp;
		return true;
	}
	return false;
}

static bool SetDebugDelayLPT (const char* delayStr, void*)
{
	if (ParseInt32 (delayStr, s_lptOptions.debugDelayMS))
		return true;

	printerr ("Debug delay option needs a valid non negative value!\n");
	return false;
}

static uint32 s_txTimeOutMS = 10*1000; // 10 second default timeout.
static bool SetTxTimeOut (const char* str, void*)
{
	if (ParseInt32 (str, s_txTimeOutMS))
		return true;

	printerr ("TX timeout option needs a valid non negative value!\n");
	return false;
}

const Option g_options[] = 
{
	{ "port",       OPT_PARAM|OPT_ONCE, SetPort },
	{ "txdelay",    OPT_PARAM|OPT_ONCE, SetTxDelayLPT },
	{ "txtimeout",  OPT_PARAM|OPT_ONCE, SetTxTimeOut },
	{ "debugdelay", OPT_PARAM|OPT_ONCE, SetDebugDelayLPT },
	{ "console",    OPT_ONCE,           SetConsole },
};

struct RamImage
{
	uint32 fileSize;
	void*  pData;
} mainRamImage = { 0 }, subRamImage = { 0 };

// Load file from disk into memory and store the file size.
bool LoadRamImage (RamImage& dest, const char* filename)
{
	// Load images from disk into memory.
	HANDLE hMain = CreateFile (filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hMain == INVALID_HANDLE_VALUE)
	{
		printerr ("Binary image '%s' not found!\n", filename);
		return false;
	}

	LARGE_INTEGER ramSize;
	if (!GetFileSizeEx (hMain, &ramSize) || ramSize.QuadPart > 32*1024)
	{
		printerr ("Invalid size for binary image '%s'!\n", filename);
		CloseHandle (hMain);
		return false;
	}

	dest.pData = malloc ((size_t)ramSize.QuadPart);
	dest.fileSize = (uint32)ramSize.QuadPart;
	DWORD bytesRead;
	if (!ReadFile (hMain, dest.pData, (uint32)dest.fileSize, &bytesRead, NULL) || bytesRead != (uint32)dest.fileSize)
	{
		printerr ("Error reading binary image '%s'!\n", filename);
		free (dest.pData);
		CloseHandle (hMain);
		return false;
	}

	CloseHandle (hMain);
	return true;
}

bool HandleInputFile (const char* arg, void* context)
{
	RamImage& dest = mainRamImage.pData ? subRamImage : mainRamImage;
	if (dest.pData)
	{
		// Both have already been loaded.
		printerr ("Invalid argument '%s'\n", arg);
		return false;
	}

	return LoadRamImage (dest, arg);
}

int main (int argc, char** argv)
{
	printf ("Out Run Bootloader %s.\n", versionString);

	int optionRes = ParseOptions (argc, argv, HandleInputFile, nullptr, g_options, sizeof(g_options)/sizeof(Option));
	if (optionRes < 0)
		return 1; // Can't parse.

	if (optionRes < 1 || !mainRamImage.pData)
	{
		// Print options.
		printf ("Usage: orboot [-options] main.bin [sub.bin]\n");
		printf ("Options: -port        Set the LPT port number (default: 0x378).\n");
		printf ("         -txdelay     Sets the strobe delay in nanoseconds (default: 50000).\n");
		printf ("         -debugdelay  Sets a debug delay in milliseconds between transitions.\n");
		printf ("         -txtimeout   Sets a timeout delay in milliseconds (0=disabled).\n");
		printf ("         -console     Keeps the console open and prints nibble mode output.\n");
		return 1;
	}

	// Start uploading.
	/*
	if (s_serialPort)
	{
		char portName[20];
		sprintf (portName, "COM%u", s_serialPort);
		g_pComm = new CommSerial (portName);
	}
	else */
	{
		// Some fallback in case we didn't specify an explicit LPT port.
		if (!s_lptPortDefined)
		{
			printf ("LPT port not explicitly specified, detecting...\n");
			LPT_PORT_INFO* portInfo = nullptr;
			size_t portsFound = LPT_DetectPorts (&portInfo);
			if (!portsFound)
			{
				printwarn ("No LPT ports detected, trying 0x%x as fallback.\n", s_lptOptions.basePort);
			}
			else
			{
				// Try to find anything with the name 'LPT1'.
				size_t portIdx = 0;
				for (size_t i=0; i<portsFound; i++)
				{
					if (portInfo[i].name && _wcsicmp (portInfo[i].name, L"LPT1") == 0)
					{
						portIdx = i;
						break;
					}
				}

				const LPT_PORT_INFO* port = portInfo+portIdx;
				if (port->portAddr)
				{
					printf ("Using port '%S' at address 0x%x\n", port->name ? port->name : L"<Unknown>", port->portAddr);
					s_lptOptions.basePort = port->portAddr;
				}
			}

			LPT_FreePortInfo (portInfo, portsFound);
		}

		g_pComm = new CommLPT (s_lptOptions);
	//	COMM_SetControlInversionMask (CONTROL_nAUTOFEED_i); // We have autofeed inverted on the PCB.
	}

	g_pComm->SetTXTimeOutMS (s_txTimeOutMS); // 10 second default timeout.

	printf ("Initializing...");
	bool bError = false;
	if (g_pComm->Init ())
	{
		g_pComm->Reset ();

		if (!Nop())
		{
			printerr ("\nBootloader device not in default state. Operation timed out.\n");
			bError = true;
		}
	}
	else
	{
		printerr ("\nError initializing output port.\n");
		//if (!s_serialPort)
			printerr ("Did you run InpOut32's InstallDriver.exe?\n"); // Needed on Vista and above.
		bError = true;
	}

	// Start uploading.
	if (!bError)
	{
		printf ("\nUploading to main ram (%u bytes)...", mainRamImage.fileSize);
		if (!UploadMainRam (mainRamImage.pData, (uint16)mainRamImage.fileSize))
		{
			printerr ("\nUpload to main ram timed out.\n");
			bError = true;
		}
		else
		{
			printf ("\n");
			if (subRamImage.pData)
			{
				printf ("Uploading to sub ram (%u bytes)...", subRamImage.fileSize);
				if (!UploadSubRam (subRamImage.pData, (uint16)subRamImage.fileSize))
				{
					printerr ("\nUpload to sub ram timed out.\n");
					bError = true;
				}
				else printf ("\n");
			}
		}
	}

	// Free loaded data.
	free (mainRamImage.pData);
	if (subRamImage.pData)
		free (subRamImage.pData);

	if (bError)
		return 1;

	printf ("Rebooting to ram...");
	if (!RebootRAM ())
	{
		printerr ("\nReboot timed out.\n");
		return 1;
	}
	printf ("\n");

	if (s_bConsole)
	{
		// Handle console output.
		printf ("\n");
		SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_LOWEST);

		for (;;)
		{
			int16 byte = g_pComm->RecvByte ();
			if (byte == -1)
			{
				printerr ("\n\nRead timed out. Shouldn't happen.\n");
				return 1;
			}
			putchar (byte);
		}
	}

	return 0;
}
