#include "platform.h"
#include "optionparse.h"
#include "error.h"
#include <string>
#include <set>

// Handles multiple options in the option description.
static bool MatchOption (const char* arg, const Option& option)
{
	const char* srch = option.name;
	for (;;)
	{
		const char* next = strchr (srch, ',');
		std::string opt (srch, next ? next-srch : strlen(srch));
		if (_stricmp (opt.c_str(), arg) == 0)
			return true;

		if (next)
			srch = next+1;
		else break;
	}

	return false;
}

int ParseOptions (int argc, const char* const* argv, ARGCALLBACK* pfnArg, void* context, const Option* pOptions, size_t optionCount)
{
	typedef std::set<const Option*> OptionSet;
	OptionSet onceOpts;

	int argCount = 0;
	for (int i=1; i<argc; i++)
	{
		const char* arg = argv[i];
		if (arg[0] == '-' || arg[0] == '/')
		{
			// It's an option.
			std::string strArg (arg+1);

			std::string::size_type sepPos = strArg.find_first_of (":="); // Note: strValue can still be empty if '-opt=' specified.
			std::string strValue;
			if (sepPos != std::string::npos)
			{
				strValue = strArg.substr (sepPos+1);
				strArg = strArg.substr (0, sepPos);
			}

			bool bFound = false;
			for (size_t j=0; j<optionCount; j++)
			{
				const Option& option = pOptions[j];
				if (MatchOption (strArg.c_str(), option))
				{
					// Once options.
					if (option.flags & OPT_ONCE)
					{
						std::pair<OptionSet::iterator,bool> insResult = onceOpts.insert (&option);
						if (!insResult.second)
						{
							printerr ("Error: option '%c%s' can appear only once!\n", arg[0], strArg.c_str());
							return -1;
						}
					}

					// Match.
					bool bSucceeded = false;
					if (!(option.flags & OPT_PARAM))
					{
						if (sepPos != std::string::npos)
						{
							printerr ("Error: option '%c%s' doesn't expect a parameter!\n", arg[0], strArg.c_str());
							return -1;
						}

						bSucceeded = option.callback (NULL, option.context);
					}
					else
					{
						// We need an extra param.
						if (sepPos == std::string::npos)
						{
							// Grab the next argument from the command line.
							if (i+1 == argc)
							{
								printerr ("Error: option '%c%s' expects a value.\n", arg[0], strArg.c_str());
								return -1;
							}

							strValue = argv[++i];
						}

						// Remove quotes.
						if (!strValue.empty())
						{
							char c = strValue[0];
							if (c == '\'' || c == '\"')
							{
								// Find trailing quote. It may not be in the current arg.
								std::string::size_type searchStartPos = 0;
								strValue = strValue.substr (1);

								for (;;)
								{
									std::string::size_type endQuote = strValue.find (c, searchStartPos);
									if (endQuote == strValue.length()-1)
									{
										// All done.
										strValue = strValue.substr (0, strValue.length()-1);
										break;
									}
									else if (endQuote != std::string::npos)
									{
										printerr ("Error: parse error inside quoted parameter of option '%c%s'.\n", arg[0], strArg.c_str());
										return -1;
									}
									else
									{
										// Glue the next argument onto this one.
										DEBUG_ASSERT(endQuote == std::string::npos);
										if (++i == argc)
										{
											printerr ("Error: option '%c%s' has an incomplete value.\n", arg[0], strArg.c_str());
											return -1;
										}

										searchStartPos = strValue.length()+1;
										strValue += ' ';
										strValue += argv[i];
									}
								} // for (;;)
							} // ' or "
						} // !strValue.Empty()

						bSucceeded = option.callback (strValue.c_str(), option.context);
					}

					if (!bSucceeded)
						return -1; // Error.

					bFound = true;
					break;
				}
			}

			if (!bFound)
				printwarn ("Warning: unknown option '%c%s'.\n", arg[0], strArg.c_str());
		}
		else
		{
			if (!pfnArg (arg, context)) // Input file?
				return -1;
			argCount++;
		}
	} // for (args)

	return argCount;
}
